"""
oas3.objects.components.link
~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

from marshmallow import fields
from oas3.base import BaseObject, BaseSchema
from oas3.objects.server import Server


class Link(BaseObject):
    """
    The Link object represents a possible design-time link for a response.
    The presence of a link does not guarantee the caller's ability to
    successfully invoke it, rather it provides a known relationship and
    traversal mechanism between responses and other operations.

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#linkObject
    """

    class Schema(BaseSchema):
        operation_ref = fields.Str(load_from='operationRef',
                                   dump_to='operationRef')
        operation_id = fields.Str(load_from='operationId',
                                  dump_to='operationId')
        parameters = fields.Dict(keys=fields.Str, values=fields.Dict)
        request_body = fields.Dict(load_from='requestBody',
                                   dump_to='requestBody')
        description = fields.Str()
        server = fields.Nested(Server.Schema)

        def represents(self):
            return Link

    def __init__(self, operation_ref=None, operation_id=None, parameters=None, request_body=None, description=None, server=None):
        self.operation_ref = operation_ref
        self.operation_id = operation_id
        self.parameters = parameters
        self.request_body = request_body
        self.description = description
        self.server = server



"""
oas3.objects.components.parameter
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

from marshmallow import fields
from oas3.base import BaseObject, BaseSchema


class Parameter(BaseObject):
    """
    Holds a set of reusable objects for different aspects of the OAS.
    All objects defined within the components object will have no effect on the API
    unless they are explicitly referenced from properties outside the components object.

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#componentsObject
    """

    class Schema(BaseSchema):
        name = fields.Str(required=True)
        location = fields.Str(required=True,
                              load_from='in',
                              dump_to='in')
        description = fields.Str()
        required = fields.Bool()
        deprecated = fields.Bool()
        allow_empty_value = fields.Bool(load_from='allowEmptyValue',
                                        dump_to='allowEmptyValue')
        style = fields.Str()
        explode = fields.Bool()
        allow_reserved = fields.Bool(load_from='allowReserved',
                                     dump_to='allowReserved')
        schema = fields.Dict()
        example = fields.Raw()
        examples = fields.Dict(keys=fields.Str, values=fields.Dict)

        def represents(self):
            return Parameter

    def __init__(self, name, location, description=None, required=None, deprecated=None, allow_empty_value=None, style=None, explode=None, allow_reserved=None, schema=None, example=None, examples=None):
        self.name = name
        self.location = location
        self.description = description
        self.required = required
        self.deprecated = deprecated
        self.allow_empty_value = allow_empty_value
        self.style = style
        self.explode = explode
        self.allow_reserved = allow_reserved
        self.schema = schema
        self.example = example
        self.examples = examples



"""
oas3.base
~~~~~~~~~
Implements base classes for internal inheritance within the API.
"""

import pathlib
import json
import yaml
import requests
import marshmallow
from inspect import cleandoc
from marshmallow import post_dump, post_load
from .errors import ValidationError, LoadingError, DumpingError


class BaseSchema(marshmallow.Schema):
    """Provides a base schema for all OAS3 object schemas to inherit from."""
    def represents(self):
        """All OAS3 object schemas must implement this method to act
        as a reference to the class the schema represents."""
        raise NotImplementedError()

    @post_dump
    def skip_none_values(self, data):
        """Skips any values that are None because they were not provided."""
        return {
            key: value for key, value in data.items()
            if value is not None
        }

    @post_load
    def make_obj(self, data):
        return self.represents()(**data)


class BaseObject:
    """Provides a base class for all OAS3 objects to inherit from."""

    @classmethod
    def from_file(cls, path):
        """
        Reads in a file from a system path to load a spec object.

        :param path: An absolute or local path to the file to be loaded
        :returns instance: Newly created object of the same type the method
            was called from

        Example:
            >>> from oas3 import Spec
            >>> spec = Spec.from_file('./tests/samples/valid/uspto.yaml')
        """
        data = open(path).read()
        extension = pathlib.Path(path).suffix
        if extension == '.json':
            return cls.from_json(data)
        if extension in ['.yaml', '.yml']:
            return cls.from_yaml(data)
        return cls.from_raw(data)

    @classmethod
    def from_url(cls, url, format_type=None):
        """
        Load a JSON or YAML OAS 3 spec object from a provided url string.

        :param url: The endpoint where the file is hosted at.
        :param format_type: either `json` or `yaml` or None, if  None it will attempt
        to be inferred.
        :returns instance: a newly created object or the type this method was called from

        Example:
            >>> from oas3 import Spec
            >>> spec = Spec.from_url('https://raw.githubusercontent.com/OAI/OpenAPI-Specification/master/examples/v3.0/petstore.yaml')
        """
        response = requests.get(url)
        if not response.ok:
            raise LoadingError('HTTP Error: {}'.format(response.status_code))
        if format_type == 'yaml':
            return cls.from_yaml(response.text)
        elif format_type == 'json':
            return cls.from_json(response.text)
        else:
            return cls.from_raw(response.text)

    @classmethod
    def from_dict(cls, dictionary):
        """
        Load the object with a python dictionary.

        :returns instance: Returns a newly created instance of the class this method
            was called from.
        :raises ValidationError: Raises if the data doesnt meet object defined schema.
        """
        result, errors = cls.Schema().load(dictionary)
        if errors:
            raise ValidationError("Validation error encountered in [{}] ".format(cls.__name__) +
                                  str(errors))
        return result

    @classmethod
    def from_json(cls, json_string):
        """
        Loads the OAS3 object with a JSON string.

        :returns instance: Returns a newly created instance of the class this method
            was called from.
        :raises ValidationError: Raises if JSON is invalid or if the specification
            data was invalid.
        """
        try:
            dictionary = json.loads(json_string)
        except:
            raise ValidationError('Unable to load, invalid JSON data')
        return cls.from_dict(dictionary)

    @classmethod
    def from_yaml(cls, yaml_string):
        """
        Loads the OAS3 object with a YAML string.

        :returns instance: Returns a newly created instance of the class this method
            was called from.
        :raises ValidationError: Raises if YAML is invalid or if the specification
            data was invalid.
        """
        try:
            dictionary = yaml.load(yaml_string)
        except:
            raise ValidationError('Unable to load, invalid YAML data')
        return cls.from_dict(dictionary)

    @classmethod
    def from_docstring(cls, obj_or_cls_or_func):
        """
        Load the OAS3 objects with the docstring of a class, object, or method.

        :returns instance: Returns a newly created instance of the class this method
            was called from.
        :raises ValidationError: Raises if data in docstring is invalid YAML or JSON,
            if the specification data was invalid, or the passed variable doesnt have
            a docstring.
        """
        if obj_or_cls_or_func.__doc__ is None:
            raise ValidationError('Attemped to load docstring but it was None')
        docstring = cleandoc(obj_or_cls_or_func.__doc__)
        return cls.from_raw(docstring)

    @classmethod
    def from_raw(cls, data):
        """
        This loader will attempt to load either JSON or YAML data, it should only
        be used if the data type is unknown at runtime, otherwise from_json() or
        from_yaml() should be used.

        :returns instance: Returns a newly created instance of the class this method
            was called from.
        :raises ValidationError: Raises if data doesnt appear to be YAML or JSON,
            or if the specification data contained is invalid.
        """
        try:
            return cls.from_yaml(data)
        except Exception as e:
            print(str(e))
        try:
            return cls.from_json(data)
        except Exception as e:
            print(str(e))
        raise ValidationError('Unable to detect valid JSON or YAML in data.')

    def to_dict(self):
        """
        Converts all subattributes to python builtin data types so that the
        OAS3 object can be represented as a dict.

        :returns dict: A dictionary representation of the OAS3 object
        :raises ValidationError: if serializing the data was unsuccessful.
        """
        data, errors = self.Schema().dump(self)
        if errors:
            raise ValidationError(errors)
        errors = self.Schema().validate(data)
        if errors:
            raise ValidationError(errors)
        return data

    def to_json(self, pretty=True, indent=2):
        """
        Converts the OAS3 object into a JSON string.

        :param pretty: If True the JSON output will be indented, sorted and nicely
            separated.
        :param indent: Ignored unless pretty=True, number of spaces to indent output
        :returns str: A JSON string of the OAS3 object
        :raises ValidationError: if serializing the data was unsuccessful.
        """
        if pretty:
            return json.dumps(self.to_dict(),
                              indent=indent,
                              sort_keys=True,
                              separators=(',', ': '))
        return json.dumps(self.to_dict())

    def to_yaml(self):
        """
        Converts the OAS3 object into a YAML string.

        :returns str: A YAML string of the OAS3 object
        :raises ValidationError: if serializing the data was unsuccessful.
        """
        return yaml.dump(self.to_dict(), default_flow_style=False)

    def to_file(self, path, format_type=None):
        file_ref = open(path, 'w')
        extension = pathlib.Path(path).suffix
        if format_type:
            extension = '.' + format_type

        if extension == '.json':
            file_ref.write(self.to_json())
            file_ref.close()
        elif extension in ['.yaml', '.yml']:
            file_ref.write(self.to_yaml())
            file_ref.close()
        else:
            raise DumpingError('Unable to determine format to save data, \
                                please specify a `format_type` if a proper file \
                                extension is not given')
        return

    def is_valid(self):
        """
        Determines if the loaded data into the OAS3 object appears to be valid,
        note that this method cannot determine the overall validity because its
        scope is limited to itself. So for example its impossible to validate
        local refs with this method.

        :returns bool: True if the object is valid in its scheme, False otherwise
        """
        data, errors = self.Schema().dump(self)
        if errors:
            return False
        errors = self.Schema().validate(data)
        if errors:
            return False
        return True



"""
oas3.objects.components.callback
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

from marshmallow import fields
from oas3.base import BaseObject, BaseSchema


class Callback(BaseObject):
    """
    Holds a set of reusable objects for different aspects of the OAS.
    All objects defined within the components object will have no effect on the API
    unless they are explicitly referenced from properties outside the components object.

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#componentsObject
    """

    class Schema(BaseSchema):
        pass

        def represents(self):
            return Callback

    def __init__(self, schemas=None):
        pass



"""
oas3.objects.components.response
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

from marshmallow import fields
from oas3.base import BaseObject, BaseSchema
from oas3.util import RefOrSchema
from oas3.objects.components.media_type import MediaType


class Response(BaseObject):
    """
    Describes a single response from an API Operation, including design-time,
    static links to operations based on the response.

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#responseObject
    """

    class Schema(BaseSchema):
        description = fields.Str(required=True)
        headers = fields.Dict()
        content = fields.Dict(keys=fields.Str, values=RefOrSchema(MediaType.Schema))
        links = fields.Dict()

        def represents(self):
            return Response

    def __init__(self, description, headers=None, content=None, links=None):
        self.description = description
        self.headers = headers
        self.content = content
        self.links = links



"""
oas3.objects.info
~~~~~~~~~~~~~~~~~
"""

from marshmallow import fields
from oas3.base import BaseObject, BaseSchema


class Contact(BaseObject):
    """
    Contact information for the exposed API.

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#contactObject
    """

    class Schema(BaseSchema):
        name = fields.Str(required=True)
        url = fields.Url(required=True)
        email = fields.Email()

        def represents(self):
            return Contact

    def __init__(self, name, url, email=None):
        self.name = name
        self.url = url
        self.email = email


class License(BaseObject):
    """
    License information for the exposed API.

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#licenseObject
    """
    class Schema(BaseSchema):
        name = fields.Str(required=True)
        url = fields.Url()

        def represents(self):
            return License

    def __init__(self, name, url=None):
        self.name = name
        self.url = url


class Info(BaseObject):
    """
    Info object provides metadata about the API.

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#infoObject
    """

    class Schema(BaseSchema):
        version = fields.Str(required=True)
        title = fields.Str(required=True)
        description = fields.Str(required=False)
        terms_of_service = fields.Str(load_from='termsOfService', dump_to='termsOfService')
        contact = fields.Nested(Contact.Schema)
        license = fields.Nested(License.Schema)

        def represents(self):
            return Info

    def __init__(self, title, version, description=None,
                 terms_of_service=None, contact=None, license=None):
        self.title = title
        self.version = version
        self.description = description
        self.terms_of_service = terms_of_service
        self.contact = contact
        self.license = license



"""
oas3.objects.components.encoding
~~~~~~~~~~~~~
"""

from marshmallow import fields
from oas3.base import BaseObject, BaseSchema


class Encoding(BaseObject):
    """
    A single encoding definition applied to a single schema property.

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#encodingObject
    """

    class Schema(BaseSchema):
        content_type = fields.Str(load_from='contentType',
                                  dump_to='contentType')
        headers = fields.Dict(keys=fields.Str, values=fields.Dict)
        style = fields.Str()
        explode = fields.Bool()
        allow_reserved = fields.Bool(load_from='allowReserved',
                                     dump_to='allowReserved')

        def represents(self):
            return Encoding

    def __init__(self, content_type=None, headers=None, style=None, explode=None, allow_reserved=None):
        self.content_type = content_type
        self.headers = headers
        self.style = style
        self.explode = explode
        self.allow_reserved = allow_reserved



"""
oas3.objects.path
~~~~~~~~~~~~~~~~~
"""

from marshmallow import fields
from oas3.base import BaseObject, BaseSchema
from oas3.objects.server import Server
from .operation import Operation
from oas3.objects.components.parameter import Parameter


class Path(BaseObject):
    """
    Adds metadata to a single tag that is used by the Operation Object.
    It is not mandatory to have a Tag Object per tag defined in the
    Operation Object instances.

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#pathItemObject
    """

    class Schema(BaseSchema):
        ref = fields.Str()
        summary = fields.Str()
        description = fields.Str()
        get = fields.Nested(Operation.Schema)
        post = fields.Nested(Operation.Schema)
        put = fields.Nested(Operation.Schema)
        patch = fields.Nested(Operation.Schema)
        delete = fields.Nested(Operation.Schema)
        options = fields.Nested(Operation.Schema)
        trace = fields.Nested(Operation.Schema)
        head = fields.Nested(Operation.Schema)
        servers = fields.List(fields.Nested(Server.Schema))
        parameters = fields.List(fields.Nested(Parameter.Schema))

        def represents(self):
            return Path

    def __init__(self, ref=None, summary=None, description=None, get=None, post=None, put=None, patch=None, delete=None, options=None, trace=None, head=None, servers=None, parameters=None):
        self.ref = ref
        self.summary = summary
        self.description = description
        self.get = get
        self.post = post
        self.put = put
        self.delete = delete
        self.options = options
        self.trace = trace
        self.head = head
        self.servers = servers
        self.parameters = parameters



"""
oas3.primitives
~~~~~~~~~~~~~~~
Defines OAS3 enumerated primitive values.
"""


class INTEGER:
    COMMON_NAME = "integer"
    TYPE = "INTEGER"
    FORMAT = "int32"


class LONG:
    COMMON_NAME = "integer"
    TYPE = "INTEGER"
    FORMAT = "int64"


class FLOAT:
    COMMON_NAME = "float"
    TYPE = "number"
    FORMAT = "float"


class DOUBLE:
    COMMON_NAME = "double"
    TYPE = "number"
    FORMAT = "double"


class STRING:
    COMMON_NAME = "string"
    TYPE = "string"
    FORMAT = None


class BYTE:
    COMMON_NAME = "byte"
    TYPE = "string"
    FORMAT = "byte"


class BINARY:
    COMMON_NAME = "binary"
    TYPE = "string"
    FORMAT = "binary"


class BOOLEAN:
    COMMON_NAME = "boolean"
    TYPE = "boolean"
    FORMAT = None


class DATE:
    COMMON_NAME = "date"
    TYPE = "string"
    FORMAT = "date"


class DATETIME:
    COMMON_NAME = "dateTime"
    TYPE = "string"
    FORMAT = "date-time"


class PASSWORD:
    COMMON_NAME = "password"
    TYPE = "string"
    FORMAT = "password"



"""
oas3.objects.tag
~~~~~~~~~~~~~~~~
"""

from marshmallow import fields
from oas3.base import BaseObject, BaseSchema
from .external_docs import ExternalDocs


class Tag(BaseObject):
    """
    Adds metadata to a single tag that is used by the Operation Object.
    It is not mandatory to have a Tag Object per tag defined in the
    Operation Object instances.

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#tagObject
    """

    class Schema(BaseSchema):
        name = fields.Str(required=True)
        description = fields.Str()
        external_docs = fields.Nested(ExternalDocs.Schema)

        def represents(self):
            return Tag

    def __init__(self, name, description=None, external_docs=None):
        self.name = name
        self.description = description
        self.external_docs = external_docs



"""
oas3.objects.components
~~~~~~~~~~~~~~~~~~~~~~~
"""

from marshmallow import fields
from oas3.base import BaseObject, BaseSchema
from .schema import Schema
from .example import Example


class Components(BaseObject):
    """
    Holds a set of reusable objects for different aspects of the OAS.
    All objects defined within the components object will have no effect on the API
    unless they are explicitly referenced from properties outside the components object.

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#componentsObject
    """

    class Schema(BaseSchema):
        example = fields.List(fields.Nested(Example.Schema))
        schemas = fields.Dict(keys=fields.Str,
                              values=fields.Nested(Schema.Schema))

        def represents(self):
            return Components

    def __init__(self, schemas=None):
        self.schemas = schemas



"""
oas3.errors
~~~~~~~~~~~
Defines custom pacakage specific exceptions.
"""


class ValidationError(Exception):
    pass


class LoadingError(Exception):
    pass


class DumpingError(Exception):
    pass



"""
oas3.objects.components.security_scheme
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

from marshmallow import fields
from oas3.base import BaseObject, BaseSchema


class OAuthFlow(BaseObject):
    """
    Configuration details for a supported OAuth Flow

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#oauthFlowObject
    """

    class Schema(BaseSchema):
        authorization_url = fields.Url(load_from='authorizationUrl',
                                       dump_to='authorizationUrl')
        token_url = fields.Url(load_from='tokenUrl',
                               dump_to='tokenUrl')
        refresh_url = fields.Url(load_from='refreshUrl',
                                 dump_to='refreshUrl')
        scopes = fields.Dict(keys=fields.Str, values=fields.Str)

        def represents(self):
            return OAuthFlow

    def __init__(self, authorization_url=None, token_url=None, refresh_url=None, scopes=None):
        self.authorization_url = authorization_url
        self.token_url = token_url
        self.refresh_url = refresh_url
        self.scopes = scopes


class OAuthFlows(BaseObject):
    """
    Allows configuration of the supported OAuth Flows.

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#oauthFlowsObject
    """

    class Schema(BaseSchema):
        implicit = fields.Nested(OAuthFlow.Schema)
        password = fields.Nested(OAuthFlow.Schema)
        client_credentials = fields.Nested(OAuthFlow.Schema,
                                           load_from='clientCredentials',
                                           dump_to='clientCredentials')
        authorization_code = fields.Nested(OAuthFlow.Schema,
                                           load_from='authorizationCode',
                                           dump_to='clientCredentials')

        def represents(self):
            return OAuthFlows

    def __init__(self, implicit=None, password=None, client_credentials=None, authorization_code=None):
        self.implicit = implicit
        self.password = password
        self.client_credentials = client_credentials
        self.authorization_code = authorization_code


class SecurityScheme(BaseObject):
    """
    Defines a security scheme that can be used by the operations.
    Supported schemes are HTTP authentication, an API key (either as a
    header or as a query parameter), OAuth2's common flows (implicit,
    password, application and access code) as defined in RFC6749, and
    OpenID Connect Discovery.

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#securitySchemeObject
    """

    class Schema(BaseSchema):
        scheme_type = fields.Str(required=True,
                                 load_from='type',
                                 dump_to='type')
        description = fields.Str()
        name = fields.Str()
        location = fields.Str()
        scheme = fields.Str()
        bearer_format = fields.Str(load_from='bearerFormat',
                                   dump_to='bearerFormat')
        flows = fields.Nested(OAuthFlows.Schema)
        open_id_connect_url = fields.Url(load_from='openIdConnectUrl',
                                         dump_to='openIdConnectUrl')

        def represents(self):
            return SecurityScheme

    def __init__(self, scheme_type, description=None, name=None, location=None, scheme=None, bearer_format=None, flows=None, open_id_connect_url=None):
        self.scheme_type = scheme_type
        self.description = description
        self.name = name
        self.location = location
        self.scheme = scheme
        self.bearer_format = bearer_format
        self.flows = flows
        self.open_id_connect_url = open_id_connect_url



"""
oas3.objects.path.operation
~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

from marshmallow import fields
from oas3.base import BaseObject, BaseSchema
from oas3.objects.external_docs import ExternalDocs
from oas3.objects.server import Server
from oas3.objects.components.response import Response
from oas3.objects.components.request_body import RequestBody
from oas3.objects.components.parameter import Parameter


class Operation(BaseObject):
    """
    Describes a single API operation on a path.

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#operationObject
    """

    class Schema(BaseSchema):
        responses = fields.Dict(required=True, keys=fields.Str, values=fields.Nested(Response.Schema))
        tags = fields.List(fields.Str)
        summary = fields.Str()
        description = fields.Str()
        external_docs = fields.Nested(ExternalDocs.Schema,
                                      load_from='externalDocs',
                                      dump_to='externalDocs')
        operation_id = fields.Str()
        parameters = fields.List(fields.Nested(Parameter.Schema))
        request_body = fields.Nested(RequestBody.Schema,
                                     load_from='requestBody',
                                     dump_to='requestBody')
        # FIXME: Use Callback.Schema
        callbacks = fields.Dict(keys=fields.Str,
                                values=fields.Dict(keys=fields.Str,
                                                   values=fields.Dict()))
        deprecated = fields.Bool()
        security = fields.List(fields.Dict(keys=fields.Str,
                                           values=fields.List(fields.Str)))
        servers = fields.List(fields.Nested(Server.Schema))

        def represents(self):
            return Operation

    def __init__(self, responses, tags=None, summary=None, description=None, external_docs=None, operation_id=None, parameters=None, request_body=None, callbacks=None, deprecated=None, security=None, servers=None):
        self.responses = responses
        self.tags = tags
        self.summary = summary
        self.description = description
        self.external_docs = external_docs
        self.operation_id = operation_id
        self.parameters = parameters
        self.request_body = request_body
        self.callbacks = callbacks
        self.deprecated = deprecated
        self.security = security
        self.servers = servers



"""
oas3.objects.components.header
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

from marshmallow import fields
from oas3.base import BaseObject, BaseSchema


class Header(BaseObject):
    """

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#headerObject
    """

    class Schema(BaseSchema):
        pass

        def represents(self):
            return Header

    def __init__(self):
        pass



"""
oas3.objects.server
~~~~~~~~~~~~~~~~~~~
"""

from marshmallow import fields
from oas3.base import BaseObject, BaseSchema


class ServerVariable(BaseObject):
    """
    An object representing a Server Variable for server URL template substitution.

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#serverVariableObject
    """

    class Schema(BaseSchema):
        default = fields.Str(required=True)
        enum = fields.List(fields.Str)
        description = fields.Str()

        def represents(self):
            return ServerVariable

    def __init__(self, default, enum=None, description=None):
        self.default = default
        self.enum = enum
        self.description = description


class Server(BaseObject):
    """
    An object representing a Server.

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#serverObject
    """

    class Schema(BaseSchema):
        url = fields.Str(required=True)
        description = fields.Url()
        variables = fields.Dict(keys=fields.Str,
                                values=fields.Nested(ServerVariable.Schema))

        def represents(self):
            return Server

    def __init__(self, url, description=None, variables=None):
        self.url = url
        self.description = description
        self.variables = variables



"""
oas3.objects.external_docs
~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

from marshmallow import fields
from oas3.base import BaseObject, BaseSchema


class ExternalDocs(BaseObject):
    """
    Allows referencing an external resource for extended documentation.

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#externalDocumentationObject
    """

    class Schema(BaseSchema):
        url = fields.Url(required=True)
        description = fields.Str()

        def represents(self):
            return ExternalDocs

    def __init__(self, url, description=None):
        self.url = url
        self.description = description



"""
oas3.objects.components.media_type
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

from marshmallow import fields
from oas3.util import RefOrSchema
from oas3.base import BaseObject, BaseSchema
from .schema import Schema


class MediaType(BaseObject):
    """
    Each Media Type Object provides schema and examples for the media
    type identified by its key.

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#mediaTypeObject
    """

    class Schema(BaseSchema):
        schema = RefOrSchema(Schema.Schema)
        example = fields.Raw()
        examples = fields.Dict(keys=fields.Str, values=fields.Dict)
        encoding = fields.Dict(keys=fields.Str, values=fields.Dict)

        def represents(self):
            return MediaType

    def __init__(self, schema=None, example=None, examples=None, encoding=None):
        self.schema = schema
        self.example = example
        self.examples = examples
        self.encoding = encoding



"""
oas3.objects.components.request_body
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

from marshmallow import fields
from oas3.base import BaseObject, BaseSchema


class RequestBody(BaseObject):
    """
    Describes a single request body.

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#requestBodyObject
    """

    class Schema(BaseSchema):
        content = fields.Dict(required=True, keys=fields.Str, values=fields.Dict)
        description = fields.Str()
        required = fields.Bool()

        def represents(self):
            return RequestBody

    def __init__(self, content, description=None, required=None):
        self.content = content
        self.description = description
        self.required = required



"""
oas3.objects.components.example
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

from marshmallow import fields
from oas3.base import BaseObject, BaseSchema


class Example(BaseObject):
    """

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#exampleObject
    """

    class Schema(BaseSchema):
        summary = fields.Str()
        description = fields.Str()
        value = fields.Raw()

        def represents(self):
            return Example

    def __init__(self, summary=None, description=None, value=None):
        self.summary = summary
        self.description = description
        self.value = value



"""
oas3.util
~~~~~~~~~
Utility functions within the library.
"""

from marshmallow.fields import Field


class RefOrSchema(Field):
    """Represents a field that should contain either a JSON reference or an Object."""

    def __init__(self, schema, **kwargs):
        super(RefOrSchema, self).__init__(**kwargs)
        self.schema = schema

    def _deserialize(self, value, attr, data):
        print(value, attr, data)



"""
oas3
~~~~
OAS3 is a package that makes compiling, parsing, validating and converting
Open API v3 specifications simple and pythonic. The library aims to be useful
as a core library for creating higher level OAS3 packages.
"""

# Versioneer
# ----------
from marshmallow import fields
from .base import BaseObject, BaseSchema
from .objects.info import Info
from .objects.server import Server
from .objects.tag import Tag
from .objects.external_docs import ExternalDocs
from .objects.components import Components, Schema
from .objects.path import Path, Operation
from .errors import LoadingError, DumpingError, ValidationError  # NOQA
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions


class Spec(BaseObject):
    """
    High level interface around compiling, validating, parsing and loading an OAS3 spec.
    """

    class Schema(BaseSchema):
        openapi = fields.Str(required=True)  # FIXME: should use default value
        info = fields.Nested(Info.Schema, required=True)
        paths = fields.Dict(required=True, keys=fields.Str, values=Path.Schema)
        servers = fields.List(fields.Nested(Server.Schema))
        components = fields.Nested(Components.Schema)
        security = fields.List(fields.Dict(keys=fields.Str,
                                           values=fields.List(fields.Str)))
        tags = fields.List(fields.Nested(Tag.Schema))
        external_docs = fields.Nested(ExternalDocs.Schema,
                                      load_from='externalDocs',
                                      dump_to='externalDocs')

        def represents(self):
            return Spec

    def __init__(self,
                 openapi=None,
                 info=None,
                 paths=None,
                 servers=None,
                 components=None,
                 security=None,
                 tags=None,
                 external_docs=None):
        self.openapi = openapi
        self.info = info
        self.paths = paths
        self.servers = servers
        self.components = components
        self.security = security
        self.tags = tags
        self.external_docs = external_docs

    def to_dict(self):
        """
        Converts all internal data type to raw dictionaries with
        built-in values.

        :returns dict: Dictionary representing the spec
        :raises ValidationError: Spec was incomplete or had errors
        """
        data, errors = self.Schema().dump(self)
        for key, value in self.paths.items():
            if isinstance(value, Path):
                self.paths[key] = value.to_dict()
        for key, value in self.components.schemas.items():
            if isinstance(value, Schema):
                self.components.schemas[key] = value.to_dict()
        if errors:
            raise ValidationError(errors)
        errors = self.Schema().validate(data)
        if errors:
            raise ValidationError(errors)
        return data



"""
setup.py for oas3
"""

__author__ = "Pinn Technologies, Inc."
__email__ = "developers@pinn.ai"
__copyright__ = "Copyright 2018, Pinn Technologies, Inc."

import os
from setuptools import setup, find_packages

this_dir = os.path.abspath(os.path.dirname(__file__))
REQUIREMENTS = filter(None, open(
    os.path.join(this_dir, 'requirements', 'main.txt')).read().splitlines())

import versioneer

setup(
    name='oas3',
    author=__author__,
    author_email=__email__,
    license="MIT",
    url="https://oas3.readthedocs.org",
    zip_safe=False,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    install_requires=list(REQUIREMENTS),
    classifiers=[k for k in open('CLASSIFIERS').read().split('\n') if k],
    description='A library for dealing with OpenAPI v3 specifications in Python',
    long_description=open('README.rst').read() + open('HISTORY.rst').read(),
    packages=find_packages(exclude=["tests*"])
)



# -*- coding: utf-8 -*-
#

import sys
import os

sys.path.insert(0, os.path.abspath('../../'))
sys.path.append(os.path.abspath('_themes'))
import oas3

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
]

templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'
project = u'OAS3'
copyright = u'2018, Pinn Technologies, Inc'

version = release = oas3.__version__
exclude_patterns = []
pygments_style = 'sphinx'
html_theme_options = {
    "index_logo": "logo.png"
}
html_theme_path = ["_themes"]
html_theme = 'flask'
html_static_path = ['_static']
#html_style = 'limiter.css'

htmlhelp_basename = 'oas3doc'
#html_logo = 'small-logo.png'
#html_favicon = 'tap-icon.png'
html_sidebars = {
    'index': ['sidebarintro.html', 'localtoc.html', 'sourcelink.html', 'searchbox.html'],
    '**': ['localtoc.html', 'relations.html',
           'sourcelink.html', 'searchbox.html']
}

latex_documents = [
    ('index', 'OAS3.tex', u'OAS3 Documentation',
     u'Pinn Technologies, Inc.', 'manual'),
]
man_pages = [
    ('index', 'OAS3.text', u'OAS3 Documentation',
     [u'Pinn Technologies, Inc.'], 1)
]

texinfo_documents = [
    ('index', 'OAS3', u'OAS3 Documentation',
     u'Pinn Technologies, Inc.', 'OAS3', 'OAS3 Python Library.',
     'Miscellaneous'),
]

intersphinx_mapping = {'python': ('http://docs.python.org/', None)
}

autodoc_default_flags = [
    "members"
    , "show-inheritance"
]



"""
oas3.objects.components.schema
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

from marshmallow import fields
from oas3.base import BaseObject, BaseSchema


class Schema(BaseObject):
    """
    The Schema Object allows the definition of input and output data types.
    These types can be objects, but also primitives and arrays. This object is
    an extended subset of the JSON Schema Specification Wright Draft 00.

    .. note:
        https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.0.md#schemaObject
    """

    class Schema(BaseSchema):
        #title =
        #multiple_of =
        #maximum =
        #exclusive_maximum =
        #minimum =
        #exclusive_minimum =
        #max_length =
        #min_length =
        #pattern =
        #max_items =
        #min_items =
        #unique_items =
        #max_properties =
        #min_properties =
        required = fields.List(fields.Str)
        #enum =
        schema_type = fields.Str(load_from='type', dump_to='type')
        all_of = fields.List(fields.Raw, load_from='allOf', dump_to='allOf')
        #one_of =
        #any_of =
        #not_of =
        items = fields.Dict()
        properties = fields.Dict()
        #additional_properties =
        #description =
        #schema_format =
        #default =
        #nullable =
        #read_only =
        #write_only =
        #xml =
        #external_docs =
        example = fields.Dict()
        #deprecated =

        def represents(self):
            return Schema

    def __init__(self, properties=None, required=None, schema_type=None, all_of=None, example=None, items=None):
        self.properties = properties
        self.required = required
        self.schema_type = schema_type
        self.all_of = all_of
        self.example = example
        self.items = items
